extern crate winres;
extern crate which;

use std::process::Command;

fn main() {
    // Build Windows icon file
    if cfg!(target_os = "windows") {
        let mut res = winres::WindowsResource::new();
        res.set_icon("resources/till_counter.ico");
        res.compile().unwrap();
    }
    // Revision Info
    set_version_env_vars()
}

/// Sets version environment vars from the Version Control System
///
/// Env Vars:
///
/// COMMIT_DATE
///
/// COMMIT_HASH
fn set_version_env_vars() {
    /// Supported Version Control Systems
    #[derive(Debug, PartialEq, Eq)]
    enum VCS {
        Git,
        Mercurial,
        Pijul,
        Fossil,
        /// Fallback for no version control
        NoVCS,
    }

    impl VCS {
        /// Returns the executable name for the VCS system
        fn binary_name(&self) -> &str {
            match self {
                Self::Git       => "git",
                Self::Mercurial => "hg",
                Self::Pijul     => "pijul",
                Self::Fossil    => "fossil",
                Self::NoVCS     => ""
            }
        }
    }

    fn vcs_from_str(s: &str) -> VCS {
        match s {
            ".git"      => VCS::Git,
            ".hg"       => VCS::Mercurial,
            ".pijul"    => VCS::Pijul,
            ".fossil"   => VCS::Fossil,
            _           => VCS::NoVCS,
        }
    }

    use which::which;
    use std::fs::read_dir;
    /////////////////////////////////////////////////////////////////////////
    // Do the needful here.
    /////////////////////////////////////////////////////////////////////////
    // Figure out what Version Control System we're using.
    // Get all directories in the project root
    let dirs = read_dir(env!("CARGO_MANIFEST_DIR")).unwrap();
    let mut vcs = VCS::NoVCS; // Default if we don't find any VCS folders.
    // Search until we find a VCS folder.
    for folder in dirs {
        let v = vcs_from_str(
                folder
                .unwrap()
                .file_name()
                .into_string()
                .unwrap()
                .as_str()
            );
        if v != VCS::NoVCS {
            vcs = v;
            break; // Go with the first VCS folder we find.
        }
    }

    // Variables to pass into env
    let mut commit_date = "0000-00-00".to_string();   // Defaults
    let mut commit_hash = "ffffffffffff".to_string(); // Defaults

    // Replace the default values below with the correct versions from VCS
    match vcs {
        VCS::NoVCS => {
            // We can't pull any information, use the Defaults assigned above.
            println!("cargo:warning=Unable to detect Version Control System, program wil not have correct build info!");
        }
        VCS::Git => {
            todo!("Git version information not implemented!!")
        }
        VCS::Mercurial => {
            let bin: &str = vcs.binary_name();
            let foo = which(bin);
            match foo {
                Ok(_val) => {
                    let stdout = String::from_utf8(
                        Command::new(bin)
                        .args(&["log", "--rev", ".", "--template", "{date|shortdate}\n{node|short}"])
                        .output()
                        .unwrap()
                        .stdout
                    )
                    .unwrap();
                    let output = stdout.lines().collect::<Vec<&str>>();
                    commit_date = output[0].trim().to_owned();
                    commit_hash = output[1].trim().to_owned();
                }
                Err(_err) => {
                    println!("cargo:warning=Attempt to use Mercurial for version information, but binary does not exist.");
                }
            }
        }
        VCS::Pijul => {
            todo!("Pijul version information not implemented!!")
        }
        VCS::Fossil => {
            todo!("Fossil version information not implemented!!")
        }
    }
    // Output the build information from version control to Cargo envs
    println!("cargo:rustc-env=COMMIT_DATE={}", commit_date);
    println!("cargo:rustc-env=COMMIT_HASH={}", commit_hash);
}