use std::io;
use std::io::Write;
use std::convert::TryFrom;


// Globals are declared outside all other scopes.
//const NAME: &'static str = env!("CARGO_PKG_NAME");
const VERSION: &'static str = env!("CARGO_PKG_VERSION");
const AUTHORS: &'static str = env!("CARGO_PKG_AUTHORS");
const REVISION: &'static str = env!("COMMIT_HASH");
const REVDATE: &'static str = env!("COMMIT_DATE");
const TERM_WIDTH: &'static usize = &40;

struct Tender {
    #[allow(dead_code)]
    name: String,
    short_name: String,
    value: u64, // Value in pennies
}

fn build_tender(name: &str, short_name: &str, value: u64) -> Tender {
    Tender {
        name: name.to_string(),
        short_name: short_name.to_string(),
        value,
    }
}

fn build_bills() ->  Vec<Tender> {
    let mut denominations : Vec<Tender> = Vec::new();
    denominations.push(build_tender("Hundreds", "$100", 10_000));
    denominations.push(build_tender("Fifties", "$50", 5_000));
    denominations.push(build_tender("Twenties", "$20", 2_000));
    denominations.push(build_tender("Tens", "$10", 1_000));
    denominations.push(build_tender("Fives", "$5", 500));
    denominations.push(build_tender("Ones", "$1", 100));
    denominations
}

fn build_coins() ->  Vec<Tender> {
    let mut denominations : Vec<Tender> = Vec::new();
    denominations.push(build_tender("Quarters", "¢25", 25));
    denominations.push(build_tender("Dimes", "¢10", 10));
    denominations.push(build_tender("Nickles", "¢5", 5));
    denominations.push(build_tender("Pennies", "¢1", 1));
    denominations
}

fn get_u32_input(prompt: &str) -> u32 {
    loop {
        print!("{}:\t", prompt); // Re-print every loop, since a bad entry makes a newline
        io::stdout().flush().expect("Couldn't flush stdout?"); // Makes the input prompts actually show in the loop
        match get_value_stdin().parse::<u32>() {
            Ok(i) => {
                return i
            },
            Err(_e) => continue,
        }
    }
}

fn get_value_stdin() -> String {
    let mut input_text = String::new();
    io::stdin()
        .read_line(&mut input_text)
        .expect("failed to read from stdin");

    input_text.trim().to_string()
}

// Text Formatting

fn display_banner(instr: &str, padding_char: Option<char>) {
    let in_chars = instr.chars().count();
    let mut output = String::new();
    let padding_char = padding_char.unwrap_or('=');

    let padding_required = TERM_WIDTH - in_chars;
    let padding_half = padding_required / 2;

    if in_chars == 0 { // If an empty string just print our padding character across the whole length
        for _n in 0..padding_required {
            output.push(padding_char);
        }
    } else { // Otherwise, we need to do some spacing stuff.
        for n in 0..padding_half {
            // Replace the final character with a space.
            if n == padding_half - 1 {
                output.push(' ')
    
            } else {
                output.push(padding_char)
            }
        }
        output.push_str(instr);
        for n in 0..padding_half {
            if n == 0 {
                output.push(' ')
            } else {
                output.push(padding_char)
            }
        }
        if output.chars().count() < padding_required { // Rounding issues, append another char
            output.push(padding_char);
        }
    }
    println!("{}", output);
}

fn display_value(value_name: &str, value: &str) {
    println!("{}\t{}", value_name, value)
}

fn format_cash(total: &u64) -> String {
    let thousands = total / 100_000;
    let dollars = (total / 100) - (thousands * 1_000);
    let cents = total - (dollars * 100) - (thousands * 100_000);
    if thousands > 0 {
        format!("${},{:03}.{:02}", thousands, dollars, cents)

    } else {
        format!("${}.{:02}", dollars, cents)
    }
}

fn display_total(total: &u64) {
    println!("{}", format_cash(total))
}

fn build_total(bills: &Vec<u32>, bill_types: &Vec<Tender>, coins: &Vec<u32>, coin_types: &Vec<Tender>) -> u64{
    let mut total = 0;
    for n in 0..bills.len() {
        let val = u64::from(bills[n]) * (bill_types[n].value);
        total += u64::from(val);
    }
    for n in 0..coins.len() {
        let val = u64::from(coins[n]) * (coin_types[n].value);
        total += u64::from(val);
    }
    total
}

fn main() {
        // Entry point only used to exit cleanly.
        let exit_code = real_main();
        std::process::exit(exit_code);
}

fn real_main() -> i32 {
    #[allow(unused_mut)]
    let mut exit_code: i32 = 0; // Default OK!

    // Who are we, what do we do?
    display_banner("Till Counter", None);
    display_value("Version: ", VERSION);
    display_value("Author:  ", AUTHORS);
    display_value("Build:  ", REVISION);
    display_value("Date:\t", REVDATE);
    display_banner("", None);

    'main_loop: loop {
        // Bills
        let bill_types = build_bills();
        let mut bills : Vec<u32> = Vec::new();
        // Count up our bills
        println!("- Bills");
        for val in bill_types.iter() {
            bills.push(get_u32_input(&val.short_name));
        }
        println!("\n");

        // Coins
        let coin_types = build_coins();
        let mut coins : Vec<u32> = Vec::new();
        //Count our coins
        println!("- Coins");
        for val in coin_types.iter() {
            coins.push(get_u32_input(&val.short_name));
        }
        println!("\n");
        print!("Cash On Hand:\t");

        // Build our total
        display_total(&build_total(&bills, &bill_types, &coins, &coin_types));
        println!("\n");

        // Action Prompt
        loop {
            println!("[Q]uit, [E]dit, [R]estart, [T]otals");
            print!("> ");
            io::stdout().flush().expect("Couldn't flush stdout?"); // Makes the input prompts actually show in the loop
            match get_value_stdin().chars().next().unwrap_or(' ').to_lowercase().next().unwrap_or(' ') {
                't' => {
                    println!("");
                    for n in 0..bills.len() {
                        let bill_total = u64::from(bills[n]) * bill_types[n].value;
                        println!("{}:\t{}\t - {}", 
                            bill_types[n].short_name, 
                            bills[n], 
                            format_cash(&bill_total)
                        );
                    }
                    println!("");
                    for n in 0..coins.len() {
                        let coin_total = u64::from(coins[n]) * coin_types[n].value;
                        println!("{}:\t{}\t - {}", 
                            coin_types[n].short_name, 
                            coins[n],
                            format_cash(&coin_total)
                        );
                    }
                    println!("");
                    print!("Cash On Hand:\t");
                    display_total(&build_total(&bills, &bill_types, &coins, &coin_types));
                    println!("\n");
                    continue;
                }
                'q' => {
                    break 'main_loop; // We're done here.
                }
                'r' => {
                    continue 'main_loop; // Starts at the beginning.
                }
                'e' => {
                    let coin_start_offset = bills.len();
                    for n in 0..bills.len() {
                        println!("[{}]\t- {}", n, bill_types[n].short_name);
                    }
                    for n in 0..coins.len() {
                        println!("[{}]\t- {}", coin_start_offset + n, coin_types[n].short_name);
                    }
                    // Get Selection
                    let mut selected: u32;
                    loop {
                        selected = get_u32_input("Currency number to edit");
                        if selected >= u32::try_from(coins.len() + coin_start_offset).unwrap() {
                            println!("Not valid!");
                            continue;
                        }
                        break
                    }
                    // We're indexing DIRECLTY into a vector here, so 'selected' MUST be valid, or we'll panic and crash.
                    if selected < u32::try_from(coin_start_offset).unwrap() {
                        bills[selected as usize] = get_u32_input(&bill_types[selected as usize].short_name); // DANGER - Direct indexing!
                    } else {
                        let selected_offset = selected - u32::try_from(coin_start_offset).unwrap();
                        coins[selected_offset as usize] = get_u32_input(&coin_types[selected_offset as usize].short_name); // DANGER - Direct indexing!
                    }
                    continue;
                }
                _ => {
                    continue;
                },
            }
        }
    }

    // Quit!
    exit_code
}
